# SPDX-License-Identifier: LGPL-2.1-or-later

import platform

from .visitor import XDRVisitor
from .parser import (
    XDRTypeString,
    XDRTypeVoid,
    XDRTypeOpaque,
    XDRTypeCustom,
    XDRDefinitionTypedef,
    XDRDeclarationFixedArray,
    XDRDeclarationVariableArray,
    XDRDeclarationPointer,
)


class XDRTypeDeclarationGenerator(XDRVisitor):
    def visit_definition_cescape(self, obj, indent, context):
        return obj.code + "\n"

    def visit_definition_constant(self, obj, indent, context):
        return "#%sdefine %s %s\n" % (indent, obj.name, obj.value)

    def visit_definition_enum(self, obj, indent, context):
        code = "%senum %s %s;\n" % (
            indent,
            obj.name,
            self.visit_object(obj.body, indent),
        ) + "%stypedef enum %s %s;\n" % (indent, obj.name, obj.name)
        return code

    def generate_cleanup(self, name, indent):
        code = "%svoid xdr_%s_clear(%s *objp);\n" % (
            indent,
            name,
            name,
        ) + "%sG_DEFINE_AUTO_CLEANUP_CLEAR_FUNC(%s, xdr_%s_clear);\n" % (
            indent,
            name,
            name,
        )
        return code

    def visit_definition_struct(self, obj, indent, context):
        code = (
            "%sstruct %s %s;\n"
            % (indent, obj.name, self.visit_object(obj.body, indent))
            + "%stypedef struct %s %s;\n" % (indent, obj.name, obj.name)
            + self.generate_cleanup(obj.name, indent)
        )
        return code

    def visit_definition_union(self, obj, indent, context):
        code = (
            "%sstruct %s %s;\n"
            % (indent, obj.name, self.visit_object(obj.body, indent, obj.name))
            + "%stypedef struct %s %s;\n" % (indent, obj.name, obj.name)
            + self.generate_cleanup(obj.name, indent)
        )
        return code

    def visit_definition_typedef(self, obj, indent, context):
        code = "%stypedef %s;\n" % (
            indent,
            self.visit_object(obj.decl, indent),
        ) + self.generate_cleanup(obj.decl.identifier, indent)
        return code

    def visit_declaration_scalar(self, obj, indent, context):
        return "%s %s" % (self.visit_object(obj.typ, indent), obj.identifier)

    def visit_declaration_pointer(self, obj, indent, context):
        return "%s *%s" % (self.visit_object(obj.typ, indent), obj.identifier)

    def visit_declaration_fixedarray(self, obj, indent, context):
        return "%s %s[%s]" % (
            self.visit_object(obj.typ, indent),
            obj.identifier,
            obj.length,
        )

    def visit_declaration_variablearray(self, obj, indent, context):
        if type(obj.typ) is XDRTypeString:
            return "%schar *%s" % (indent, obj.identifier)
        else:
            code = (
                "%sstruct {\n" % indent
                + "%s    u_int %s_len;\n" % (indent, obj.identifier)
                + "%s    %s *%s_val;\n"
                % (indent, self.visit_object(obj.typ, ""), obj.identifier)
                + "%s} %s" % (indent, obj.identifier)
            )
            return code

    def visit_type_custom(self, obj, indent, context):
        return "%s%s" % (indent, obj.identifier)

    def visit_type_opaque(self, obj, indent, context):
        return "%schar" % indent

    def visit_type_string(self, obj, indent, context):
        return "%sstring" % indent

    def visit_type_void(self, obj, indent, context):
        return "%svoid" % indent

    def visit_type_char(self, obj, indent, context):
        return "%schar" % indent

    def visit_type_unsignedchar(self, obj, indent, context):
        return "%su_char" % indent

    def visit_type_short(self, obj, indent, context):
        return "%sshort" % indent

    def visit_type_unsignedshort(self, obj, indent, context):
        return "%su_short" % indent

    def visit_type_int(self, obj, indent, context):
        return "%sint" % indent

    def visit_type_unsignedint(self, obj, indent, context):
        return "%su_int" % indent

    def visit_type_hyper(self, obj, indent, context):
        return "%sint64_t" % indent

    def visit_type_unsignedhyper(self, obj, indent, context):
        return "%suint64_t" % indent

    def visit_type_bool(self, obj, indent, context):
        return "%sbool_t" % indent

    def visit_type_float(self, obj, indent, context):
        return "%sfloat" % indent

    def visit_type_double(self, obj, indent, context):
        return "%sdouble" % indent

    def visit_type_enum(self, obj, indent, context):
        return "%senum %s" % (indent, self.visit_object(obj.body.body, indent))

    def visit_type_struct(self, obj, indent, context):
        return "%sstruct %s" % (indent, self.visit_object(obj.body, indent))

    def visit_type_union(self, obj, indent, context):
        return "%sstruct %s" % (indent, self.visit_object(obj.body, indent))

    def visit_enum_value(self, obj, indent, context):
        return "%s%s = %s" % (indent, obj.name, obj.value)

    def visit_enum_body(self, obj, indent, context):
        code = "{\n"
        for value in obj.values:
            code = code + self.visit_object(value, indent + "    ") + ",\n"
        code = code + "%s}" % indent
        return code

    def visit_struct_body(self, obj, indent, context):
        code = "{\n"
        for value in obj.fields:
            code = code + self.visit_object(value, indent + "    ") + ";\n"
        code = code + "%s}" % indent
        return code

    def visit_union_case(self, obj, indent, context):
        return self.visit_object(obj.decl, indent)

    def visit_union_body(self, obj, indent, context):
        prefix = context
        if prefix != "":
            prefix = prefix + "_"

        code = (
            "%s{\n" % indent
            + "%s    %s;\n" % (indent, self.visit_object(obj.discriminator))
            + "%s    union {\n" % indent
        )
        for value in obj.cases:
            if type(value.decl.typ) is XDRTypeVoid:
                continue
            code = code + self.visit_object(value, indent + "        ") + ";\n"
        if obj.default is not None and type(obj.default.typ) is not XDRTypeVoid:
            code = code + self.visit_object(obj.default, indent + "        ") + ";\n"
        code = code + "%s    } %su;\n" % (indent, prefix) + "%s}" % indent
        return code


class XDRTypeImplementationGenerator(XDRVisitor):
    def visit_definition_enum(self, obj, indent, context):
        pass

    def generate_cleanup(self, name, indent):
        code = (
            "\n"
            + "%svoid xdr_%s_clear(%s *objp)\n" % (indent, name, name)
            + "%s{\n" % indent
            + "%s    xdr_free((xdrproc_t)xdr_%s, (char *)objp);\n" % (indent, name)
            + "%s}\n" % indent
        )
        return code

    def visit_definition_union(self, obj, indent, context):
        return self.generate_cleanup(obj.name, indent)

    def visit_definition_struct(self, obj, indent, context):
        return self.generate_cleanup(obj.name, indent)

    def visit_definition_typedef(self, obj, indent, context):
        return self.generate_cleanup(obj.decl.identifier, indent)


class XDRMarshallDeclarationGenerator(XDRVisitor):
    def visit_definition_enum(self, obj, indent, context):
        return "%sextern  bool_t xdr_%s(XDR *, %s*);\n" % (indent, obj.name, obj.name)

    def visit_definition_union(self, obj, indent, context):
        return "%sextern  bool_t xdr_%s(XDR *, %s*);\n" % (indent, obj.name, obj.name)

    def visit_definition_struct(self, obj, indent, context):
        return "%sextern  bool_t xdr_%s(XDR *, %s*);\n" % (indent, obj.name, obj.name)

    def visit_definition_typedef(self, obj, indent, context):
        if isinstance(obj.decl, XDRDeclarationFixedArray):
            return "%sextern  bool_t xdr_%s(XDR *, %s);\n" % (
                indent,
                obj.decl.identifier,
                obj.decl.identifier,
            )
        else:
            return "%sextern  bool_t xdr_%s(XDR *, %s*);\n" % (
                indent,
                obj.decl.identifier,
                obj.decl.identifier,
            )


class XDRMarshallImplementationGenerator(XDRVisitor):
    def visit_definition_enum(self, obj, indent, context):
        code = (
            "%sbool_t\n" % indent
            + "%sxdr_%s(XDR *xdrs, %s *objp)\n" % (indent, obj.name, obj.name)
            + "%s{\n" % indent
            + "%s    if (!xdr_enum(xdrs, (enum_t *)objp))\n" % indent
            + "%s        return FALSE;\n" % indent
            + "%s    return TRUE;\n" % indent
            + "%s}\n" % indent
        )
        return code

    def generate_type_call(self, decl, field, typename, embedded=False, indent=""):
        if type(decl.typ) is XDRTypeVoid:
            return ""
        if type(decl) is XDRDeclarationFixedArray:
            if type(decl.typ) is XDRTypeOpaque:
                code = "%s    if (!xdr_%s(xdrs, %s, %s))\n" % (
                    indent,
                    self.visit_object(decl.typ, context="func"),
                    field,
                    decl.length,
                )
            else:
                code = "%s    if (!xdr_vector(xdrs, (char *)%s, %s,\n" % (
                    indent,
                    field,
                    decl.length,
                ) + "%s        sizeof(%s), (xdrproc_t)xdr_%s))\n" % (
                    indent,
                    self.visit_object(decl.typ),
                    self.visit_object(decl.typ, context="func"),
                )
        elif type(decl) is XDRDeclarationVariableArray:
            fieldRef = "."
            pointerStr = ""
            if embedded:
                pointerStr = "&"
            else:
                fieldRef = "->"

            if type(decl.typ) is XDRTypeString:
                code = "%s    if (!xdr_%s(xdrs, %s%s, %s))\n" % (
                    indent,
                    self.visit_object(decl.typ, context="func"),
                    pointerStr,
                    field,
                    decl.maxlength,
                )
            elif type(decl.typ) is XDRTypeOpaque:
                code = "%s    if (!xdr_bytes(xdrs, (char **)&%s%s%s_val, " % (
                    indent,
                    field,
                    fieldRef,
                    typename,
                ) + "(u_int *) &%s%s%s_len, %s))\n" % (
                    field,
                    fieldRef,
                    typename,
                    decl.maxlength,
                )
            else:
                code = (
                    "%s    if (!xdr_array(xdrs, (char **)&%s%s%s_val, "
                    % (indent, field, fieldRef, typename)
                    + "(u_int *) &%s%s%s_len, %s,\n"
                    % (field, fieldRef, typename, decl.maxlength)
                    + "%s        sizeof(%s), (xdrproc_t)xdr_%s))\n"
                    % (
                        indent,
                        self.visit_object(decl.typ),
                        self.visit_object(decl.typ, context="func"),
                    )
                )
        elif type(decl) is XDRDeclarationPointer:
            pointerStr = ""
            if embedded:
                pointerStr = "&"

            code = "%s    if (!xdr_pointer(xdrs, (char **)%s%s, " % (
                indent,
                pointerStr,
                field,
            ) + "sizeof(%s), (xdrproc_t)xdr_%s))\n" % (
                self.visit_object(decl.typ, context="func"),
                self.visit_object(decl.typ, context="func"),
            )
        else:
            pointerStr = ""
            isFixedArray = (
                type(decl.typ) is XDRTypeCustom
                and type(decl.typ.definition) is XDRDefinitionTypedef
                and type(decl.typ.definition.decl) is XDRDeclarationFixedArray
            )

            if embedded and not isFixedArray:
                pointerStr = "&"

            code = "%s    if (!xdr_%s(xdrs, %s%s))\n" % (
                indent,
                self.visit_object(decl.typ, context="func"),
                pointerStr,
                field,
            )

        code = code + "%s        return FALSE;\n" % indent
        return code

    def visit_definition_union(self, obj, indent, context):
        code = (
            "%sbool_t\n" % indent
            + "%sxdr_%s(XDR *xdrs, %s *objp)\n" % (indent, obj.name, obj.name)
            + "%s{\n" % indent
            + self.generate_type_call(
                obj.body.discriminator,
                "objp->%s" % obj.body.discriminator.identifier,
                obj.body.discriminator.identifier,
                embedded=True,
                indent=indent,
            )
            + "%s    switch (objp->%s) {\n"
            % (indent, obj.body.discriminator.identifier)
        )

        for case in obj.body.cases:
            code = (
                code
                + "%s    case %s:\n" % (indent, case.value)
                + self.generate_type_call(
                    case.decl,
                    "objp->%s_u.%s" % (obj.name, case.decl.identifier),
                    obj.name,
                    embedded=True,
                    indent=indent + "    ",
                )
                + "%s        break;\n" % indent
            )

        code = code + "%s    default:\n" % indent

        if obj.body.default is not None:
            code = (
                code
                + self.generate_type_call(
                    obj.body.default,
                    "objp->%s_u.%s" % (obj.name, obj.body.default.identifier),
                    obj.name,
                    embedded=True,
                    indent=indent + "    ",
                )
                + "%s        break;\n" % indent
            )
        else:
            code = code + "%s        return FALSE;\n" % indent

        code = (
            code
            + "%s    }\n" % indent
            + "%s    return TRUE;\n" % indent
            + "%s}\n" % indent
        )
        return code

    def visit_definition_struct(self, obj, indent, context):
        code = (
            "%sbool_t\n" % indent
            + "%sxdr_%s(XDR *xdrs, %s *objp)\n" % (indent, obj.name, obj.name)
            + "%s{\n" % indent
        )
        for field in obj.body.fields:
            code = code + self.generate_type_call(
                field,
                "objp->%s" % field.identifier,
                field.identifier,
                embedded=True,
                indent=indent,
            )
        code = code + "%s    return TRUE;\n" % indent + "%s}\n" % indent
        return code

    def visit_definition_typedef(self, obj, indent, context):
        code = "%sbool_t\n" % indent
        if isinstance(obj.decl, XDRDeclarationFixedArray):
            code = code + "%sxdr_%s(XDR *xdrs, %s objp)\n" % (
                indent,
                obj.decl.identifier,
                obj.decl.identifier,
            )
        else:
            code = code + "%sxdr_%s(XDR *xdrs, %s *objp)\n" % (
                indent,
                obj.decl.identifier,
                obj.decl.identifier,
            )
        code = (
            code
            + "%s{\n" % indent
            + self.generate_type_call(
                obj.decl, "objp", obj.decl.identifier, embedded=False, indent=indent
            )
            + "%s    return TRUE;\n" % indent
            + "%s}\n" % indent
        )
        return code

    def visit_declaration_pointer(self, obj, indent, context):
        return "%s%s *%s" % (indent, self.visit_object(obj.typ), obj.identifier)

    def visit_declaration_fixedarray(self, obj, indent, context):
        return "%s%s %s[%s]" % (
            indent,
            self.visit_object(obj.typ),
            obj.identifier,
            obj.length,
        )

    def visit_declaration_variablearray(self, obj, indent, context):
        return "%s%s *%s" % (indent, self.visit_object(obj.typ), obj.identifier)

    def visit_type_custom(self, obj, indent, context):
        return "%s%s" % (indent, obj.identifier)

    def visit_type_opaque(self, obj, indent, context):
        return "%sopaque" % indent

    def visit_type_string(self, obj, indent, context):
        return "%sstring" % indent

    def visit_type_char(self, obj, indent, context):
        return "%schar" % indent

    def visit_type_unsignedchar(self, obj, indent, context):
        return "%su_char" % indent

    def visit_type_short(self, obj, indent, context):
        return "%sshort" % indent

    def visit_type_unsignedshort(self, obj, indent, context):
        return "%su_short" % indent

    def visit_type_int(self, obj, indent, context):
        return "%sint" % indent

    def visit_type_unsignedint(self, obj, indent, context):
        return "%su_int" % indent

    def visit_type_hyper(self, obj, indent, context):
        return "%sint64_t" % indent

    def visit_type_unsignedhyper(self, obj, indent, context):
        if context == "func" and platform.system() == "Darwin":
            return "%su_int64_t" % indent
        else:
            return "%suint64_t" % indent

    def visit_type_bool(self, obj, indent, context):
        if context == "func":
            return "%sbool" % indent
        else:
            return "%sbool_t" % indent

    def visit_type_float(self, obj, indent, context):
        return "%sfloat" % indent

    def visit_type_double(self, obj, indent, context):
        return "%sdouble" % indent

    def visit_enum_value(self, obj, indent, context):
        return "%s%s = %s" % (indent, obj.name, obj.value)

    def visit_union_case(self, obj, indent, context):
        return self.visit_object(obj.value, indent)



# SPDX-License-Identifier: LGPL-2.1-or-later

import os
from pathlib import Path

from rpcgen.parser import XDRParser
from rpcgen.generator import (
    XDRTypeDeclarationGenerator,
    XDRTypeImplementationGenerator,
    XDRMarshallDeclarationGenerator,
    XDRMarshallImplementationGenerator,
)


def test_generate_header():
    x = Path(Path(__file__).parent, "demo.x")
    h = Path(Path(__file__).parent, "demo.h")
    with x.open("r") as fp:
        parser = XDRParser(fp)
        spec = parser.parse()

    got = (
        XDRTypeDeclarationGenerator(spec).visit()
        + "\n"
        + XDRMarshallDeclarationGenerator(spec).visit()
    )

    with h.open("r") as fp:
        want = fp.read()

    if "VIR_TEST_REGENERATE_OUTPUT" in os.environ:
        want = got
        with h.open("w") as fp:
            fp.write(want)

    assert got == want


def test_generate_source():
    x = Path(Path(__file__).parent, "demo.x")
    h = Path(Path(__file__).parent, "demo.c")
    with x.open("r") as fp:
        parser = XDRParser(fp)
        spec = parser.parse()

    got = (
        XDRTypeImplementationGenerator(spec).visit()
        + "\n"
        + XDRMarshallImplementationGenerator(spec).visit()
    )

    with h.open("r") as fp:
        want = fp.read()

    if "VIR_TEST_REGENERATE_OUTPUT" in os.environ:
        want = got
        with h.open("w") as fp:
            fp.write(want)

    assert got == want



#!/usr/bin/env python3

import argparse
import enum
import os
import re

try:
    import pycpuinfo
except ImportError:
    pycpuinfo = None


# features in qemu that we do not want in libvirt
FEATURES_IGNORE = (
    "kvm-asyncpf",
    "kvm-asyncpf-int",
    "kvm-asyncpf-vmexit",
    "kvm-hint-dedicated",
    "kvm-mmu",
    "kvm-msi-ext-dest-id",
    "kvm-nopiodelay",
    "kvm-poll-control",
    "kvm-pv-eoi",
    "kvm-pv-ipi",
    "kvm-pv-sched-yield",
    "kvm-pv-tlb-flush",
    "kvm-pv-unhalt",
    "kvm-steal-time",
    "kvmclock",
    "kvmclock-stable-bit",
    "kvmclock2",

    "full-width-write",
)


# features in libvirt, that qemu does not know. as python cannot use dicts
# as keys in other dicts, use tuples. three-tuples "eax, ecx, register name"
# for cpuid features; one-tuples "index" for msrs. The values for the dict are
# mappings from "bit index" to "feature name".
FEATURES_EXTRA = {
    (0x00000001, None, "ecx"): {
        27: "osxsave",
    },
    (0x00000007, 0x0000, "ebx"): {
        12: "cmt",
    },
    (0x00000007, 0x0000, "ecx"): {
        4: "ospke",
    },
    (0x00000007, 0x0000, "edx"): {
        18: "pconfig",
    },
    (0x0000000f, 0x0001, "edx"): {
        1: "mbm_total",
        2: "mbm_local",
    },
    (0x80000001, None, "ecx"): {
        18: "cvt16",
    },
    (0x0000048c,): {
        8: "vmx-ept-uc",
        14: "vmx-ept-wb",
        41: "vmx-invvpid-single-context",               # wrong name in qemu
        43: "vmx-invvpid-single-context-noglobals",     # wrong name in qemu
    }
}


# alias information to add to generated file
FEATURES_ALIASES = {
    "arch-capabilities": (
        ("arch_capabilities", "linux"),
    ),
    "cmp_legacy": (
        ("cmp-legacy", "qemu"),
    ),
    "cmt": (
        ("cqm", "linux"),
    ),
    "ds_cpl": (
        ("ds-cpl", "qemu"),
    ),
    "fxsr_opt": (
        ("ffxsr", "qemu"),
        ("fxsr-opt", "qemu"),
    ),
    "lahf_lm": (
        ("lahf-lm", "qemu"),
    ),
    "lm": (
        ("i64", "qemu"),
    ),
    "md-clear": (
        ("md_clear", "linux"),
    ),
    "nodeid_msr": (
        ("nodeid-msr", "qemu"),
    ),
    "nrip-save": (
        ("nrip_save", "qemu"),
    ),
    "nx": (
        ("xd", "qemu"),
    ),
    "pause-filter": (
        ("pause_filter", "qemu"),
    ),
    "pclmuldq": (
        ("pclmulqdq", "qemu"),
    ),
    "perfctr_core": (
        ("perfctr-core", "qemu"),
    ),
    "perfctr_nb": (
        ("perfctr-nb", "qemu"),
    ),
    "pni": (
        ("sse3", "qemu"),
    ),
    "sse4.1": (
        ("sse4-1", "qemu"),
        ("sse4_1", "qemu"),
    ),
    "sse4.2": (
        ("sse4-2", "qemu"),
        ("sse4_2", "qemu"),
    ),
    "svm-lock": (
        ("svm_lock", "qemu"),
    ),
    "tsc-scale": (
        ("tsc_scale", "qemu"),
    ),
    "tsc_adjust": (
        ("tsc-adjust", "qemu"),
    ),
    "vmcb-clean": (
        ("vmcb_clean", "qemu"),
    ),
    "vmx-invvpid-single-context-noglobals": (
        ("vmx-invept-single-context-noglobals", "qemu"),
    ),
}


# list non-migratable features here
FEATURES_NON_MIGRATABLE = (
    "xsaves",
    "invtsc",
)


# mapping from "symbol name" to "value" for "#define"s in qemu source code
_CONSTANTS = dict()


# tree of known features. top level index is either "cpuid" or "msr".
# further indices for cpuid: eax_in, ecx_in (may be `None`), register name
# further indices for msr: index
_FEATURES = dict()


# fill _CONSTANTS with the #defines from qemu source code
def read_headers(path):
    pattern_define = re.compile("^#define\\s+(\\S+)\\s+(.*)$")

    headers = (
        "include/standard-headers/asm-x86/kvm_para.h",
        "target/i386/cpu.h",
    )

    _CONSTANTS["true"] = "1"

    for header in headers:
        with open(os.path.join(path, header), "tr") as f:
            for line in f.readlines():
                match = pattern_define.match(line)
                if match:
                    key = match.group(1)
                    val = match.group(2)
                    _CONSTANTS[key] = val


# add new cpuid feature bit
def add_feature_cpuid(eax, ecx, reg, bit, name):
    if not name:
        return

    if "cpuid" not in _FEATURES:
        _FEATURES["cpuid"] = dict()

    if eax not in _FEATURES["cpuid"]:
        _FEATURES["cpuid"][eax] = dict()

    if ecx not in _FEATURES["cpuid"][eax]:
        _FEATURES["cpuid"][eax][ecx] = dict()

    if reg not in _FEATURES["cpuid"][eax][ecx]:
        _FEATURES["cpuid"][eax][ecx][reg] = dict()

    _FEATURES["cpuid"][eax][ecx][reg][bit] = name


class VmxMsr(enum.Enum):
    MSR_IA32_VMX_PROCBASED_CTLS2 = 0x0000048b
    MSR_IA32_VMX_TRUE_PINBASED_CTLS = 0x0000048d
    MSR_IA32_VMX_TRUE_PROCBASED_CTLS = 0x0000048e
    MSR_IA32_VMX_TRUE_ENTRY_CTLS = 0x00000490
    MSR_IA32_VMX_TRUE_EXIT_CTLS = 0x0000048f


def is_vmx_msr(msr):
    try:
        VmxMsr(msr)
        return True
    except ValueError:
        return False


# add new msr feature bit
def add_feature_msr(msr, bit, name):
    if not name:
        return

    if "msr" not in _FEATURES:
        _FEATURES["msr"] = dict()

    if msr not in _FEATURES["msr"]:
        _FEATURES["msr"][msr] = dict()

    # VMX-* features are specified in the 32 higher bits
    # of the MSR value
    if is_vmx_msr(msr):
        bit += 32

    _FEATURES["msr"][msr][bit] = name


# add features from EXTRA_FEATURE to the list of known features
def add_extra_features():
    for key, val in FEATURES_EXTRA.items():
        for bit, name in val.items():
            if len(key) == 3:
                add_feature_cpuid(key[0], key[1], key[2], bit, name)
            else:
                add_feature_msr(key[0], bit, name)


# add a feature from qemu to the list of known features. translates features
# names according to FEATURE_ALIASES and applies symbolic values defined in
# _CONSTANTS.
def add_feature_qemu(query, data):
    # split names into individual items
    data = [n.strip() for n in "".join(data).split(",")]

    names = dict()
    if any([e.startswith("[") for e in data]):
        for entry in data:
            entry = entry.strip()
            if not entry:
                continue
            index, name = entry.split("=", 2)
            index = int(index.strip().strip("[").strip("]"), 0)
            names[index] = name.strip().strip("\"")
    else:
        for index, name in enumerate(data):
            if not name or name == "NULL":
                continue
            name = name.strip("\"")
            if name in FEATURES_IGNORE:
                continue
            names[index] = name.strip("\"")

    # cut out part between "{" and "}". easiest way to get rid of unwanted
    # extra info such as ".tcg_features" or multi line comments
    query = "".join(query).split("{")[1].split("}")[0]

    eax = None
    ecx = None
    reg = None
    msr = None
    for entry in [e.strip() for e in query.split(",")]:
        if not entry:
            continue
        left, right = [e.strip() for e in entry.split("=", 2)]
        if left == ".eax":
            eax = int(_CONSTANTS.get(right, right), 0)
        if left == ".ecx":
            ecx = int(_CONSTANTS.get(right, right), 0)
        if left == ".reg":
            reg = right.lower()[2:]
        if left == ".index":
            msr = int(_CONSTANTS.get(right, right), 0)

    # qemu defines some empty feature words, filter them out
    if not names:
        return
    if all([e is None for e in names.values()]):
        return

    # apply name translation and add to list of known features
    for bit, name in sorted(names.items()):
        for newname, data in FEATURES_ALIASES.items():
            for oldname, source in data:
                if name == oldname and source == "qemu":
                    name = newname

        if msr:
            add_feature_msr(msr, bit, name)
        else:
            add_feature_cpuid(eax, ecx, reg, bit, name)


def add_features_cpuinfo():
    def decode_bit(value):
        for i in range(0, 64):
            if value == (1 << i):
                return i

    def decode_cpuid(v):
        if v[0] != 0 and v[1] == 0 and v[2] == 0 and v[3] == 0:
            reg, val = "eax", v[0]
        if v[0] == 0 and v[1] != 0 and v[2] == 0 and v[3] == 0:
            reg, val = "ebx", v[1]
        if v[0] == 0 and v[1] == 0 and v[2] != 0 and v[3] == 0:
            reg, val = "ecx", v[2]
        if v[0] == 0 and v[1] == 0 and v[2] == 0 and v[3] != 0:
            reg, val = "edx", v[3]

        return reg, decode_bit(val)

    x86 = pycpuinfo.Family.find("x86", "")

    for feature in pycpuinfo.features():
        if feature.family() != x86:
            continue

        if list(feature.features()):
            continue

        name = feature.name("libvirt")
        if name in FEATURES_IGNORE:
            continue

        cpuid = feature.extra_x86_cpuid()
        if cpuid:
            eax = cpuid[0]
            ecx = cpuid[1]
            if ecx == pycpuinfo.x86.CPUINFO_X86_CPUID_ECX_NONE:
                ecx = None
            reg, bit = decode_cpuid(cpuid[2:])
            add_feature_cpuid(eax, ecx, reg, bit, name)

        msr = feature.extra_x86_msr()
        if msr:
            index = msr[0]
            bit = decode_bit(msr[1] | (msr[2] << 32))
            add_feature_msr(index, bit, name)


# read the `feature_word_info` struct from qemu's cpu.c into a list of strings
def read_cpu_c(path):
    pattern_comment = re.compile("/\\*.*?\\*/")
    marker_begin = "FeatureWordInfo feature_word_info[FEATURE_WORDS] = {\n"
    marker_end = "};\n"

    with open(os.path.join(path, "target/i386/cpu.c"), "tr") as f:
        # skip until begin marker
        while True:
            line = f.readline()
            if not line:
                exit("begin marker not found in cpu.c")
            if line == marker_begin:
                break

        # read until end marker
        while True:
            line = f.readline()
            if not line:
                exit("end marker not found in cpu.c")
            if line == marker_end:
                break

            # remove comments and white space
            line = re.sub(pattern_comment, "", line).strip()

            yield line


# simple state machine to extract feature names and definitions from extracted
# qemu source code
def parse_feature_words(lines):
    state_waiting_for_type = 1
    state_waiting_for_names = 2
    state_read_names = 3
    state_waiting_for_query = 4
    state_read_query = 5

    pattern_type = re.compile("^\\.type\\s*=\\s*(.+)$")
    pattern_names = re.compile("^\\.feat_names\\s*=\\s*{$")
    pattern_data = re.compile("^\\.(cpuid|msr).*$")
    pattern_end = re.compile("^},?$")

    state = state_waiting_for_type
    for line in lines:
        if state == state_waiting_for_type:
            match = pattern_type.match(line)
            if match:
                data_names = list()
                data_query = list()
                state = state_waiting_for_names

        elif state == state_waiting_for_names:
            # special case for missing ".feat_names" entry:
            match = pattern_data.match(line)
            if match:
                data_query.append(line)
                state = state_read_query
                continue

            match = pattern_names.match(line)
            if match:
                state = state_read_names

        elif state == state_read_names:
            match = pattern_end.match(line)
            if match:
                state = state_waiting_for_query
            else:
                data_names.append(line)

        elif state == state_waiting_for_query:
            match = pattern_data.match(line)
            if match:
                data_query.append(line)
                state = state_read_query

        elif state == state_read_query:
            match = pattern_end.match(line)
            data_query.append(line)
            if match:
                state = state_waiting_for_type
                add_feature_qemu(data_query, data_names)

        else:
            exit("parsing state machine in invalid state")

    if state != state_waiting_for_type:
        exit("parsing incomplete")


# generate x86_features.xml from list of known features
def write_output(path):
    with open(path, "tw") as f:
        f.write("<!--\n  Generated file, do not edit!\n  Use the ")
        f.write("sync_qemu_features_i386.py script to make changes.\n-->\n\n")
        f.write("<cpus>\n")

        for eax in sorted(_FEATURES["cpuid"]):
            for ecx in sorted(_FEATURES["cpuid"][eax]):
                for reg in sorted(_FEATURES["cpuid"][eax][ecx]):
                    f.write(f"\n  <!-- cpuid level 0x{eax:08x}")
                    if ecx is not None:
                        f.write(f", 0x{ecx:04x}")
                    f.write(f" ({reg:s}) -->\n")

                    names = sorted(_FEATURES["cpuid"][eax][ecx][reg].items())
                    for bit, name in names:
                        mask = 1 << bit
                        f.write(f"  <feature name='{name}'")
                        if name in FEATURES_NON_MIGRATABLE:
                            f.write(" migratable='no'")
                        f.write(">\n")
                        for alias in FEATURES_ALIASES.get(name, []):
                            f.write(f"    <alias name='{alias[0]}'")
                            f.write(f" source='{alias[1]}'/>\n")
                        f.write(f"    <cpuid eax_in='0x{eax:08x}' ")
                        if ecx is not None:
                            f.write(f"ecx_in='0x{ecx:08x}' ")
                        f.write(f"{reg:s}='0x{mask:08x}'/>\n")
                        f.write("  </feature>\n")

        for msr in sorted(_FEATURES["msr"]):
            f.write(f"\n  <!-- msr 0x{msr:08x} -->\n")
            names = sorted(_FEATURES["msr"][msr].items())
            for bit, name in names:
                mask = 1 << bit
                f.write(f"  <feature name='{name}'")
                if name in FEATURES_NON_MIGRATABLE:
                    f.write(" migratable='no'")
                f.write(">\n")
                for alias in FEATURES_ALIASES.get(name, []):
                    f.write(f"    <alias name='{alias[0]}'")
                    f.write(f" source='{alias[1]}'/>\n")
                f.write(f"    <msr index='0x{msr:08x}' ")
                f.write(f"edx='0x{(mask >> 32):08x}' ")
                f.write(f"eax='0x{(mask & 0xffffffff):08x}'/>\n")
                f.write("  </feature>\n")

        f.write("</cpus>\n")


def main():
    dirname = os.path.dirname(__file__)

    parser = argparse.ArgumentParser(
        description="Synchronize x86 cpu features from QEMU."
    )
    parser.add_argument(
        "qemu",
        help="Path to qemu source code",
        default=os.path.realpath(os.path.join(dirname, "../../../qemu")),
        nargs="?",
        type=os.path.realpath,
    )
    if pycpuinfo:
        parser.add_argument(
            "--libcpuinfo",
            help="Use libcpuinfo as data source instead",
            action="store_true",
        )
    parser.add_argument(
        "--output",
        "-o",
        help="Path to output file",
        default=os.path.realpath(os.path.join(dirname, "x86_features.xml")),
        type=os.path.realpath
    )
    args = parser.parse_args()

    if pycpuinfo and args.libcpuinfo:
        add_features_cpuinfo()
    else:
        if not os.path.isdir(args.qemu):
            parser.print_help()
            exit("qemu source directory not found")

        read_headers(args.qemu)
        lines = read_cpu_c(args.qemu)
        parse_feature_words(lines)
        add_extra_features()

    write_output(args.output)

    print(
        "After adding new features, update existing test files by running "
        "`tests/cputestdata/cpu-data.py diff tests/cputestdata/"
        "x86_64-cpuid-*.json`"
    )


if __name__ == "__main__":
    exit(main())



#!/usr/bin/env python3
# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse
import os
import sys

from rpcgen.parser import XDRParser
from rpcgen.generator import (
    XDRTypeDeclarationGenerator,
    XDRTypeImplementationGenerator,
    XDRMarshallDeclarationGenerator,
    XDRMarshallImplementationGenerator,
)


def parse_cli():
    parser = argparse.ArgumentParser("RPC code generator")
    parser.add_argument(
        "-m",
        "--mode",
        choices=["header", "source", "repr"],
        help="Output generation mode",
    )
    parser.add_argument(
        "-r", "--header", default=[], action="append", help="Extra headers to include"
    )
    parser.add_argument("input", default="-", nargs="?", help="XDR input protocol file")
    parser.add_argument("output", default="-", nargs="?", help="Generated output file")

    return parser.parse_args()


def main():
    args = parse_cli()

    infp = sys.stdin
    outfp = sys.stdout
    if args.input != "-":
        infp = open(args.input, "r")
    if args.output != "-":
        # the old genprotocol.pl wrapper would make the
        # output files mode 0444, which will prevent us
        # from writing directly do them. Explicitly
        # unlinking first gets rid of any old possibly
        # read-only copy
        #
        # We can delete this in a few years, once we
        # know users won't have a previously generated
        # readonly copy lieing around.
        try:
            os.unlink(args.output)
        except Exception:
            pass
        outfp = open(args.output, "w")

    parser = XDRParser(infp)
    spec = parser.parse()

    if args.mode == "header":
        print("/* This file is auto-generated from %s */\n" % args.input, file=outfp)
        print("#include <rpc/rpc.h>", file=outfp)
        print('#include "internal.h"', file=outfp)
        for h in args.header:
            print('#include "%s"' % h, file=outfp)
        print("", file=outfp)
        print("#pragma once\n", file=outfp)
        generator = XDRTypeDeclarationGenerator(spec)
        print(generator.visit(), file=outfp)
        generator = XDRMarshallDeclarationGenerator(spec)
        print(generator.visit(), file=outfp)
    elif args.mode == "source":
        print("/* This file is auto-generated from %s */\n" % args.input, file=outfp)
        print("#include <config.h>", file=outfp)
        for h in args.header:
            print('#include "%s"' % h, file=outfp)
        print("", file=outfp)
        generator = XDRTypeImplementationGenerator(spec)
        print(generator.visit(), file=outfp)
        generator = XDRMarshallImplementationGenerator(spec)
        print(generator.visit(), file=outfp)
    elif args.mode == "repr":
        print(spec, file=outfp)
    else:
        pass  # Just validates XDR input syntax


main()



#!/usr/bin/env python3

# Copyright (C) 2023 Red Hat, Inc.
# SPDX-License-Identifier: LGPL-2.1-or-later

import sys

SECTIONS = [
    "[Unit]",
    "[Service]",
    "[Socket]",
    "[Install]",
]


def parse_unit(unit_path):
    unit = {}
    current_section = "[Invalid]"

    with open(unit_path) as f:
        for line in f:
            line = line.strip()

            if line == "":
                continue

            # Comments at the very top of the file, outside of any
            # section, are allowed and will not show up in the output
            if line[0] == "#" and current_section == "[Invalid]":
                continue

            if line[0] == "[" and line[-1] == "]":
                if line not in SECTIONS:
                    print("Unknown section {}".format(line))
                    sys.exit(1)

                current_section = line
                continue

            if current_section not in unit:
                unit[current_section] = []

            unit[current_section].append(line)

    if "[Invalid]" in unit:
        print("Contents found outside of any section")
        sys.exit(1)

    return unit


def format_unit(unit):
    lines = []

    for section in SECTIONS:
        if section not in unit:
            continue

        lines.append(section)

        for line in unit[section]:
            lines.append(line)

        lines.append("")

    return "\n".join(lines)


def merge_units(base, extra):
    merged = {}

    for section in SECTIONS:
        if section in extra and section not in base:
            print("Section {} in extra but not in base".format(section))
            sys.exit(1)

        if section not in base:
            continue

        merged[section] = base[section]

        if section not in extra:
            continue

        merged[section].extend(extra[section])

    return merged


prog = sys.argv[0]
args = sys.argv[1:]

if len(args) < 1:
    print("usage: {} BASE [EXTRA]...".format(prog))
    sys.exit(1)

merged = parse_unit(args[0])

for arg in args[1:]:
    extra = parse_unit(arg)

    merged = merge_units(merged, extra)

sys.stdout.write(format_unit(merged))



# SPDX-License-Identifier: LGPL-2.1-or-later

from .lexer import (
    XDRLexer,
    XDRTokenPunctuation,
    XDRTokenIdentifier,
    XDRTokenCEscape,
    XDRTokenConstant,
)
from .ast import (
    XDRSpecification,
    XDRDefinitionConstant,
    XDRDefinitionTypedef,
    XDRDefinitionEnum,
    XDRDefinitionStruct,
    XDRDefinitionUnion,
    XDRDefinitionCEscape,
    XDRDeclarationScalar,
    XDRDeclarationPointer,
    XDRDeclarationFixedArray,
    XDRDeclarationVariableArray,
    XDRTypeVoid,
    XDRTypeChar,
    XDRTypeUnsignedChar,
    XDRTypeShort,
    XDRTypeUnsignedShort,
    XDRTypeInt,
    XDRTypeUnsignedInt,
    XDRTypeHyper,
    XDRTypeUnsignedHyper,
    XDRTypeFloat,
    XDRTypeDouble,
    XDRTypeBool,
    XDRTypeOpaque,
    XDRTypeString,
    XDRTypeCustom,
    XDREnumValue,
    XDREnumBody,
    XDRTypeEnum,
    XDRStructBody,
    XDRTypeStruct,
    XDRUnionCase,
    XDRUnionBody,
    XDRTypeUnion,
)


# We are parsing (approximately the following grammar
# from RFC 4506 #6.3:
#
#    declaration:
#         type-specifier identifier
#       | type-specifier identifier "[" value "]"
#       | type-specifier identifier "<" [ value ] ">"
#       | "opaque" identifier "[" value "]"
#       | "opaque" identifier "<" [ value ] ">"
#       | "string" identifier "<" [ value ] ">"
#       | type-specifier "*" identifier
#       | "void"
#
#    value:
#         constant
#       | identifier
#
#    constant:
#       decimal-constant | hexadecimal-constant | octal-constant
#
#    type-specifier:
#         [ "unsigned" ] "int"
#       | [ "unsigned" ] "hyper"
#       | "float"
#       | "double"
#       | "quadruple"     /* We're skipping this one */
#       | "bool"
#       | enum-type-spec
#       | struct-type-spec
#       | union-type-spec
#       | identifier
#
#    enum-type-spec:
#       "enum" enum-body
#
#    enum-body:
#       "{"
#          ( identifier "=" value )
#          ( "," identifier "=" value )*
#       "}"
#
#    struct-type-spec:
#       "struct" struct-body
#
#    struct-body:
#       "{"
#          ( declaration ";" )
#          ( declaration ";" )*
#       "}"
#
#    union-type-spec:
#       "union" union-body
#
#    union-body:
#       "switch" "(" declaration ")" "{"
#          case-spec
#          case-spec *
#          [ "default" ":" declaration ";" ]
#       "}"
#
#    case-spec:
#      ( "case" value ":")
#      ( "case" value ":") *
#      declaration ";"
#
#    constant-def:
#       "const" identifier "=" constant ";"
#
#    type-def:
#         "typedef" declaration ";"
#       | "enum" identifier enum-body ";"
#       | "struct" identifier struct-body ";"
#       | "union" identifier union-body ";"
#
#    definition:
#         type-def
#       | constant-def
#
#    specification:
#         definition *
#
# Notable divergance:
#
#   - In 'type-decl' we allow 'char' and 'short'
#     in signed and unsigned variants
#
#   - In 'definition' we allow '%...' as escape C code
#     to passthrough to the header output
#
#   - In 'enum-type-spec' we allow a bare enum name
#     instead of enum body
#
#   - In 'struct-type-spec' we allow a bare struct name
#     instead of struct body
#
#   - In 'union-type-spec' we allow a bare union name
#     instead of union body
#
class XDRParser:
    def __init__(self, fp):
        self.lexer = XDRLexer(fp)
        self.typedefs = {}

    def parse(self):
        spec = XDRSpecification()
        while True:
            definition = self.parse_definition()
            if definition is None:
                break
            spec.definitions.append(definition)
        return spec

    def parse_definition(self):
        token = self.lexer.next()
        if token is None:
            return None

        if type(token) is XDRTokenCEscape:
            return XDRDefinitionCEscape(token.value[1:])

        if type(token) is not XDRTokenIdentifier:
            raise Exception("Expected identifier, but got %s" % token)

        defs = {
            "const": XDRDefinitionConstant,
            "typedef": XDRDefinitionTypedef,
            "enum": XDRDefinitionEnum,
            "struct": XDRDefinitionStruct,
            "union": XDRDefinitionUnion,
        }

        if token.value not in defs:
            raise Exception("Unexpected identifier %s" % token)

        funcname = "parse_definition_" + token.value
        func = getattr(self, funcname)
        assert func is not None

        definition = func()

        semi = self.lexer.next()
        if type(semi) is not XDRTokenPunctuation or semi.value != ";":
            raise Exception("Expected ';', but got %s" % semi)

        return definition

    def parse_definition_const(self):
        ident = self.lexer.next()
        if type(ident) is not XDRTokenIdentifier:
            raise Exception("Expected identifier, but got %s" % ident)

        assign = self.lexer.next()
        if type(assign) is not XDRTokenPunctuation or assign.value != "=":
            raise Exception("Expected '=', but got %s" % assign)

        const = self.lexer.next()
        if type(const) not in [XDRTokenConstant, XDRTokenIdentifier]:
            raise Exception("Expected constant, but got %s" % const)

        return XDRDefinitionConstant(ident.value, const.value)

    def parse_definition_typedef(self):
        decl = self.parse_declaration()
        if decl.identifier in self.typedefs:
            raise Exception("Type '%s' already defined" % decl.identifier)

        definition = XDRDefinitionTypedef(decl)
        self.typedefs[decl.identifier] = definition
        return definition

    def parse_definition_enum(self):
        name = self.lexer.next()
        if type(name) is not XDRTokenIdentifier:
            raise Exception("Expected identifier, but got %s" % name)

        body = self.parse_enum_body()

        if name.value in self.typedefs:
            raise Exception("Type '%s' already defined" % name.value)

        definition = XDRDefinitionEnum(name.value, body)
        self.typedefs[name.value] = definition
        return definition

    def parse_definition_struct(self):
        name = self.lexer.next()
        if type(name) is not XDRTokenIdentifier:
            raise Exception("Expected identifier, but got %s" % name)

        body = self.parse_struct_body()

        if name.value in self.typedefs:
            raise Exception("Type '%s' already defined" % name.value)

        definition = XDRDefinitionStruct(name.value, body)
        self.typedefs[name.value] = definition
        return definition

    def parse_definition_union(self):
        name = self.lexer.next()
        if type(name) is not XDRTokenIdentifier:
            raise Exception("Expected identifier, but got %s" % name)

        body = self.parse_union_body()

        if name.value in self.typedefs:
            raise Exception("Type '%s' already defined" % name.value)

        definition = XDRDefinitionUnion(name.value, body)
        self.typedefs[name.value] = definition
        return definition

    def parse_declaration(self):
        typ = self.parse_type()

        if type(typ) is XDRTypeVoid:
            return XDRDeclarationScalar(typ, None)

        ident = self.lexer.next()

        pointer = False
        if type(ident) is XDRTokenPunctuation:
            if ident.value != "*":
                raise Exception("Expected '*' or identifer, but got %s" % ident)
            if type(typ) is XDRTypeString or type(typ) is XDRTypeOpaque:
                raise Exception("Pointer invalid for 'string' and 'opaque' types")

            pointer = True
            ident = self.lexer.next()

        bracket = self.lexer.peek()
        if type(bracket) is XDRTokenPunctuation:
            if bracket.value == "[":
                _ = self.lexer.next()
                value = self.lexer.next()
                if type(value) not in [XDRTokenConstant, XDRTokenIdentifier]:
                    raise Exception("Expected constant, but got %s" % value)

                close = self.lexer.next()
                if type(close) is not XDRTokenPunctuation or close.value != "]":
                    raise Exception("Expected ']', but got %s" % value)

                if type(typ) is XDRTypeString:
                    raise Exception("Fixed array invalid for 'string' type")
                return XDRDeclarationFixedArray(typ, ident.value, value.value)
            elif bracket.value == "<":
                _ = self.lexer.next()
                maybeValue = self.lexer.peek()
                value = None
                if type(maybeValue) in [XDRTokenConstant, XDRTokenIdentifier]:
                    value = self.lexer.next().value

                close = self.lexer.next()
                if type(close) is not XDRTokenPunctuation or close.value != ">":
                    raise Exception("Expected '>', but got %s" % close)

                return XDRDeclarationVariableArray(typ, ident.value, value)

        if pointer:
            return XDRDeclarationPointer(typ, ident.value)
        else:
            return XDRDeclarationScalar(typ, ident.value)

    def parse_type(self):
        typ = self.lexer.next()
        if type(typ) is not XDRTokenIdentifier:
            raise Exception("Expected identifier, but got %s" % typ)

        if typ.value == "unsigned":
            typ = self.lexer.peek()
            if type(typ) is not XDRTokenIdentifier:
                raise Exception("Expected identifier, but got %s" % typ)

            if typ.value == "char":
                _ = self.lexer.next()
                return XDRTypeUnsignedChar()
            elif typ.value == "short":
                _ = self.lexer.next()
                return XDRTypeUnsignedShort()
            elif typ.value == "int":
                _ = self.lexer.next()
                return XDRTypeUnsignedInt()
            elif typ.value == "hyper":
                _ = self.lexer.next()
                return XDRTypeUnsignedHyper()
            else:
                # Bare 'unsigned' isn't allowed by 'type-specifier'
                # grammer in RFC 1014, but rpcgen allows it
                return XDRTypeUnsignedInt()

        if typ.value == "void":
            return XDRTypeVoid()
        elif typ.value == "char":
            return XDRTypeChar()
        elif typ.value == "short":
            return XDRTypeShort()
        elif typ.value == "int":
            return XDRTypeInt()
        elif typ.value == "hyper":
            return XDRTypeHyper()
        elif typ.value == "float":
            return XDRTypeFloat()
        elif typ.value == "double":
            return XDRTypeDouble()
        elif typ.value == "bool":
            return XDRTypeBool()
        elif typ.value == "enum":
            return self.parse_type_enum()
        elif typ.value == "struct":
            return self.parse_type_struct()
        elif typ.value == "union":
            return self.parse_type_union()
        elif typ.value == "opaque":
            return XDRTypeOpaque()
        elif typ.value == "string":
            return XDRTypeString()
        else:
            return XDRTypeCustom(typ.value, self.typedefs.get(typ.value, None))

    def parse_enum_body(self):
        body = self.lexer.next()
        if type(body) is not XDRTokenPunctuation or body.value != "{":
            raise Exception("Expected '{', but got %s" % body)

        values = []
        while True:
            ident = self.lexer.next()
            if type(ident) is not XDRTokenIdentifier:
                raise Exception("Expected identifier, but got %s" % ident)

            equal = self.lexer.next()
            if type(equal) is not XDRTokenPunctuation or equal.value != "=":
                raise Exception("Expected '=', but got %s" % ident)

            value = self.lexer.next()
            if type(value) is not XDRTokenConstant:
                raise Exception("Expected constant, but got %s" % ident)

            separator = self.lexer.next()
            if type(separator) is not XDRTokenPunctuation and separator.value not in [
                "}",
                ",",
            ]:
                raise Exception("Expected '}' or ',', but got %s" % separator)

            values.append(XDREnumValue(ident.value, value.value))

            if separator.value == "}":
                break

        return XDREnumBody(values)

    def parse_type_enum(self):
        body = self.parse_enum_body()
        return XDRTypeEnum(body)

    def parse_struct_body(self):
        body = self.lexer.next()
        if type(body) is not XDRTokenPunctuation or body.value != "{":
            raise Exception("Expected '{', but got %s" % body)

        fields = []
        while True:
            field = self.parse_declaration()
            fields.append(field)

            separator = self.lexer.next()
            if type(separator) is not XDRTokenPunctuation and separator.value != ";":
                raise Exception("Expected ';', but got %s" % separator)

            end = self.lexer.peek()
            if type(end) is XDRTokenPunctuation and end.value == "}":
                break

        # discard the '}' we peeked at to end the loop
        _ = self.lexer.next()
        return XDRStructBody(fields)

    def parse_type_struct(self):
        body = self.parse_struct_body()
        return XDRTypeStruct(body)

    def parse_union_body(self):
        ident = self.lexer.next()
        if type(ident) is not XDRTokenIdentifier or ident.value != "switch":
            raise Exception("Expected 'switch', but got %s" % ident)

        bracket = self.lexer.next()
        if type(bracket) is not XDRTokenPunctuation or bracket.value != "(":
            raise Exception("Expected '(', but got %s" % bracket)

        discriminator = self.parse_declaration()

        bracket = self.lexer.next()
        if type(bracket) is not XDRTokenPunctuation or bracket.value != ")":
            raise Exception("Expected ')', but got %s" % bracket)

        bracket = self.lexer.next()
        if type(bracket) is not XDRTokenPunctuation or bracket.value != "{":
            raise Exception("Expected '{', but got %s" % bracket)

        default = None
        cases = []
        while True:
            ident = self.lexer.next()
            if type(ident) is not XDRTokenIdentifier or ident.value not in [
                "default",
                "case",
            ]:
                raise Exception("Expected 'default' or 'case', but got %s" % ident)

            value = None
            if ident.value == "case":
                value = self.lexer.next()
                if type(value) not in [XDRTokenConstant, XDRTokenIdentifier]:
                    raise Exception("Expected constant, but got %s" % value)

                sep = self.lexer.next()
                if type(sep) is not XDRTokenPunctuation or sep.value != ":":
                    raise Exception("Expected ':', but got %s" % value)

                decl = self.parse_declaration()

                case = XDRUnionCase(value.value, decl)
                cases.append(case)
            else:
                if default is not None:
                    raise Exception("Duplicate 'default' clause")

                sep = self.lexer.next()
                if type(sep) is not XDRTokenPunctuation or sep.value != ":":
                    raise Exception("Expected ':', but got %s" % value)

                default = self.parse_declaration()

            separator = self.lexer.next()
            if type(separator) is not XDRTokenPunctuation and separator.value != ";":
                raise Exception("Expected ';', but got %s" % bracket)

            end = self.lexer.peek()
            if type(end) is XDRTokenPunctuation and end.value == "}":
                break

        # discard the '}' we peeked at to end the loop
        _ = self.lexer.next()
        return XDRUnionBody(discriminator, cases, default)

    def parse_type_union(self):
        body = self.parse_union_body()
        return XDRTypeUnion(body)



#!/usr/bin/env python3
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see
# <http://www.gnu.org/licenses/>.
#
# Check that external references between documentation HTML files are not broken.

import argparse
import os
import re
import sys
import xml.etree.ElementTree as ET

ns = {'html': 'http://www.w3.org/1999/xhtml'}
externallinks = []
externalimages = []


def get_file_list(prefix):
    filelist = []
    imagelist = []
    imageformats = ['.jpg', '.svg', '.png']

    for root, dir, files in os.walk(prefix):
        for file in files:
            ext = os.path.splitext(file)[1]

            if ext == '.html':
                # the 404 page doesn't play well
                if '404.html' in file:
                    continue

                filelist.append(os.path.join(root, file))

            elif ext in imageformats:
                imagelist.append(os.path.join(root, file))

    filelist.sort()
    imagelist.sort()

    return filelist, imagelist


# loads an XHTML and extracts all anchors, local and remote links for the one file
def process_file(filename, project_uri):
    tree = ET.parse(filename)
    root = tree.getroot()
    docname = root.get('data-sourcedoc')
    dirname = os.path.dirname(filename)

    if not docname:
        docname = filename

    anchors = [filename]
    targets = []
    images = []
    projectlinks = []

    for elem in root.findall('.//html:a', ns):
        target = elem.get('href')
        an = elem.get('id')

        if an:
            anchors.append(filename + '#' + an)

        if target:
            if re.search('://', target):
                externallinks.append(target)

                if project_uri is not None and target.startswith(project_uri):
                    projectlinks.append((target, docname))

            elif target[0] != '#' and 'mailto:' not in target:
                targetfull = os.path.normpath(os.path.join(dirname, target))

                targets.append((filename, docname, targetfull, target))

    # older docutils generate "<div class='section'"
    for elem in root.findall('.//html:div/[@class=\'section\']', ns):
        an = elem.get('id')

        if an:
            anchors.append(filename + '#' + an)

    # modern docutils generate a <section element
    for elem in root.findall('.//html:section', ns):
        an = elem.get('id')

        if an:
            anchors.append(filename + '#' + an)

    # find local images
    for elem in root.findall('.//html:img', ns):
        src = elem.get('src')

        if src:
            if re.search('://', src):
                externalimages.append(src)
            else:
                imagefull = os.path.normpath(os.path.join(dirname, src))
                images.append((imagefull, docname))

    return (anchors, targets, images, projectlinks)


def process_all(filelist, project_uri):
    anchors = []
    targets = []
    images = []
    projectlinks = []

    for file in filelist:
        anchor, target, image, projectlink = process_file(file, project_uri)

        targets = targets + target
        anchors = anchors + anchor
        images = images + image
        projectlinks = projectlinks + projectlink

    return (targets, anchors, images, projectlinks)


def check_targets(targets, anchors):
    errors = []
    for _, docname, target, targetorig in targets:
        if target not in anchors:
            errors.append((docname, targetorig))

    if errors:
        errors.sort()

        for file, target in errors:
            print(f'ERROR: \'{file}\': broken link to: \'{target}\'')

        return True

    return False


def check_usage_crawl(page, targets, visited):
    visited.append(page)

    tocrawl = []

    for filename, docname, target, _ in targets:
        if page != filename:
            continue

        targetpage = target.split("#", 1)[0]

        if targetpage not in visited and targetpage not in tocrawl:
            tocrawl.append(targetpage)

    for crawl in tocrawl:
        check_usage_crawl(crawl, targets, visited)


# crawls the document references starting from entrypoint and tries to find
# unreachable pages
def check_usage(targets, files, entrypoint):
    visited = []
    fail = False

    check_usage_crawl(entrypoint, targets, visited)

    for file in files:
        if file not in visited:
            brokendoc = file

            for filename, docname, _, _ in targets:
                if filename != file:
                    continue
                if docname:
                    brokendoc = docname
                    break

            print(f'ERROR: \'{brokendoc}\': is not referenced from anywhere')
            fail = True

    return fail


# checks that images present in the directory are being used and also that
# pages link to existing images. For favicons, which are not referenced from
# the '.html' files there's a builtin set of exceptions.
def check_images(usedimages, imagefiles, ignoreimages):
    favicons = [
        'android-chrome-192x192.png',
        'android-chrome-256x256.png',
        'apple-touch-icon.png',
        'favicon-16x16.png',
        'favicon-32x32.png',
        'mstile-150x150.png',
    ]
    fail = False

    if ignoreimages:
        favicons = favicons + ignoreimages

    for usedimage, docname in usedimages:
        if usedimage not in imagefiles:
            print(f'ERROR: \'{docname}\' references image \'{usedimage}\' not among images')
            fail = True

    for imagefile in imagefiles:
        used = False

        if imagefile in (usedimage[0] for usedimage in usedimages):
            used = True
        else:
            for favicon in favicons:
                if favicon in imagefile:
                    used = True
                    break

        if not used:
            print(f'ERROR: Image \'{imagefile}\' is not used by any page')
            fail = True

    return fail


# checks that all links are accessed via https
def check_https(links):
    fail = False

    for link in links:
        if link.startswith('http://'):
            print(f'ERROR: URI \'{link}\' uses insecure "http" protocol')
            fail = True

    return fail


# checks prohibited external links to local files
def check_projectlinks(projectlinks, exceptions):
    fail = False

    for (link, filename) in projectlinks:
        allowed = False

        if exceptions is not None:
            for exc in exceptions:
                if exc in filename:
                    allowed = True
                    break

        if not allowed:
            print(f'ERROR: prohibited external URI \'{link}\' to local project in \'{filename}\'')
            fail = True

    return fail


parser = argparse.ArgumentParser(description='HTML reference checker')
parser.add_argument('--webroot', required=True,
                    help='path to the web root')
parser.add_argument('--entrypoint', default="index.html",
                    help='file name of web entry point relative to --webroot')
parser.add_argument('--external', action="store_true",
                    help='print external references instead')
parser.add_argument('--ignore-images', action='append',
                    help='paths to images that should be considered as used')
parser.add_argument('--require-https', action="store_true",
                    help='require secure https for external links')
parser.add_argument('--project-uri',
                    help='external prefix of the local project (e.g. https://libvirt.org; external links with that prefix are prohibited')
parser.add_argument('--project-uri-exceptions', action='append',
                    help='list of path prefixes excluded from the "--project-uri" checks')

args = parser.parse_args()

files, imagefiles = get_file_list(os.path.abspath(args.webroot))

entrypoint = os.path.join(os.path.abspath(args.webroot), args.entrypoint)

targets, anchors, usedimages, projectlinks = process_all(files, args.project_uri)

fail = False

if args.external:
    prev = None
    externallinks.sort()
    for ext in externallinks:
        if ext != prev:
            print(f'link: {ext}')

        prev = ext

    externalimages.sort()
    for ext in externalimages:
        if ext != prev:
            print(f'image: {ext}')

        prev = ext
else:
    if check_targets(targets, anchors):
        fail = True

    if check_usage(targets, files, entrypoint):
        fail = True

    if check_images(usedimages, imagefiles, args.ignore_images):
        fail = True

    if check_projectlinks(projectlinks, args.project_uri_exceptions):
        fail = True

    if args.require_https:
        if check_https(externallinks):
            fail = True

        if check_https(externalimages):
            fail = True

    if fail:
        sys.exit(1)

    sys.exit(0)



#!/usr/bin/env python3

import os
import shutil
import sys

destdir = os.environ.get('DESTDIR', os.sep)

for desc in sys.argv[1:]:
    inst = desc.split(':')
    src = inst[0]
    dst = os.path.join(destdir, inst[1].strip(os.sep))
    os.makedirs(dst, exist_ok=True)
    shutil.copy(src, dst)



# SPDX-License-Identifier: LGPL-2.1-or-later

import abc

from .ast import (
    XDRSpecification,
    XDRDefinition,
    XDRDeclaration,
    XDRType,
    XDREnumValue,
    XDREnumBody,
    XDRStructBody,
    XDRUnionCase,
    XDRUnionBody,
)


class XDRVisitor(abc.ABC):
    def __init__(self, spec):
        self.spec = spec

    def visit(self, indent="", context=""):
        return self.visit_object(self.spec, indent="", context=context)

    def visit_object(self, obj, indent="", context=""):
        if isinstance(obj, XDRSpecification):
            funcname = "visit_specification"
        elif isinstance(obj, XDRDefinition):
            funcname = "visit_definition_" + type(obj).__name__[13:].lower()
        elif isinstance(obj, XDRDeclaration):
            funcname = "visit_declaration_" + type(obj).__name__[14:].lower()
        elif isinstance(obj, XDRType):
            funcname = "visit_type_" + type(obj).__name__[7:].lower()
        elif isinstance(obj, XDREnumValue):
            funcname = "visit_enum_value"
        elif isinstance(obj, XDREnumBody):
            funcname = "visit_enum_body"
        elif isinstance(obj, XDRStructBody):
            funcname = "visit_struct_body"
        elif isinstance(obj, XDRUnionCase):
            funcname = "visit_union_case"
        elif isinstance(obj, XDRUnionBody):
            funcname = "visit_union_body"
        else:
            raise Exception("Unhandled %s" % obj.__class__.__name__)

        func = getattr(self, funcname)
        assert func is not None
        return func(obj, indent, context)

    def visit_specification(self, obj, indent, context):
        code = []
        for definition in self.spec.definitions:
            defcode = self.visit_object(definition, indent, context)
            if defcode is not None:
                code.append(defcode)
        return "\n".join(code)

    def visit_definition_cescape(self, obj, indent, context):
        pass

    def visit_definition_constant(self, obj, indent, context):
        pass

    def visit_definition_enum(self, obj, indent, context):
        pass

    def visit_definition_struct(self, obj, indent, context):
        pass

    def visit_definition_union(self, obj, indent, context):
        pass

    def visit_definition_typedef(self, obj, indent, context):
        pass

    def visit_declaration_scalar(self, obj, indent, context):
        pass

    def visit_declaration_pointer(self, obj, indent, context):
        pass

    def visit_declaration_fixedarray(self, obj, indent, context):
        pass

    def visit_declaration_variablearray(self, obj, indent, context):
        pass

    def visit_type_custom(self, obj, indent, context):
        pass

    def visit_type_opaque(self, obj, indent, context):
        pass

    def visit_type_string(self, obj, indent, context):
        pass

    def visit_type_void(self, obj, indent, context):
        pass

    def visit_type_char(self, obj, indent, context):
        pass

    def visit_type_unsignedchar(self, obj, indent, context):
        pass

    def visit_type_short(self, obj, indent, context):
        pass

    def visit_type_unsignedshort(self, obj, indent, context):
        pass

    def visit_type_int(self, obj, indent, context):
        pass

    def visit_type_unsignedint(self, obj, indent, context):
        pass

    def visit_type_hyper(self, obj, indent, context):
        pass

    def visit_type_unsignedhyper(self, obj, indent, context):
        pass

    def visit_type_bool(self, obj, indent, context):
        pass

    def visit_type_float(self, obj, indent, context):
        pass

    def visit_type_double(self, obj, indent, context):
        pass

    def visit_type_enum(self, obj, indent, context):
        pass

    def visit_type_struct(self, obj, indent, context):
        pass

    def visit_type_union(self, obj, indent, context):
        pass

    def visit_enum_value(self, obj, indent, context):
        pass

    def visit_enum_body(self, obj, indent, context):
        pass

    def visit_struct_body(self, obj, indent, context):
        pass

    def visit_union_case(self, obj, indent, context):
        pass

    def visit_union_body(self, obj, indent, context):
        pass
